# coding=utf-8
"""
Filesystem functions.
"""
from os import kill, getpid
from signal import SIGILL
import socket
import re
import time
import logging
import threading

__INITIAL_INTERVAL__ = 3
__MAX_INTERVAL_STEPS__ = 10


class DrobloPostgreSQLPoller(threading.Thread):
    """
    Thread polling our postgre server, sends a signal to droblod upon availability.
    """
    def __init__(self, connection_string, droblo_poll, sig=SIGILL, interval=__INITIAL_INTERVAL__):
        super().__init__()
        self.droblo_poll = droblo_poll
        self._steps = interval
        self._interval = interval
        self._connection_string = connection_string
        self._signal_to_send = sig
        self._server = None
        self._port = None
        self._server_reachable = True
        self._parse_connection_string()
        self._test_srv_connectivity()
        logging.info(f"DrobloPostgreSQLPoller initialized, polling {self._server}")
        logging.info(f"{self._server} is %sreacheable" % ("" if self.server_reachable else "un"))

    def run(self):
        """
        threading.Thread run override.
        """
        try:
            while self.droblo_poll.is_set():
                self._test_srv_connectivity()
                time.sleep(self.interval)
            logging.info("Exiting DrobloPostgreSQLPoller")
        except (KeyboardInterrupt, SystemExit):
            logging.info("Exiting DrobloPostgreSQLPoller")

    @property
    def interval(self) -> int:
        """
        Polling interval in seconds
        :return:
        """
        return self._interval

    @interval.setter
    def interval(self, value):
        self._interval = value
        logging.info("Polling for %s connectivity every %s seconds." % (self._server, self.interval))

    @property
    def server_reachable(self) -> bool:
        """
        Server is reachable boolean.
        :return:
        """
        return self._server_reachable

    @server_reachable.setter
    def server_reachable(self, value):
        if value != self._server_reachable:
            self._server_reachable = value
            self._send_reconnect_signal()

    def _parse_connection_string(self):
        match = re.search(r'postgresql://.*:.*@(?P<srv>.*):(?P<port>.*)/.*', self._connection_string)
        if match:
            self._server = match.group("srv")
            self._port = int(match.group("port"))

    def _test_srv_connectivity(self):
        s = socket.socket()
        try:
            s.connect((self._server, self._port))
            self.server_reachable = True
        except socket.error as _:
            if self.interval < __MAX_INTERVAL_STEPS__ * self._steps:
                self.interval += self._steps
            self.server_reachable = False
        finally:
            s.close()

    def _send_reconnect_signal(self):
        if self.server_reachable:
            logging.info("%s reachable." % self._server)
            logging.info("Resetting polling interval.")
            self.interval = self._steps
            kill(getpid(), self._signal_to_send)
        else:
            logging.warning("%s unreachable." % self._server)


def _acknowledge(signum, _):
    print("Received signal %s" % signum)


if __name__ == "__main__":
    from pap_logger import *
    from signal import signal

    PaPLogger(level=DEBUG)
    signal(SIGILL, _acknowledge)
    do_poll = threading.Event()
    do_poll.set()
    try:
        DrobloPostgreSQLPoller("postgresql://user:pass@postgresSQL:5432/db", droblo_poll=do_poll, interval=1).start()
        while True:
            print("hello")
            time.sleep(2)
    except KeyboardInterrupt:
        do_poll.clear()
        print("Bye!")



#!/usr/bin/python3
# coding=utf-8
"""
Droblod is a filesystem watcher that will take checksum of all new files, insert them in DB
"""

from pap_logger import *
import argparse
import signal
import threading
import time

from pathlib import Path
from datetime import datetime

from watchdog.observers import Observer

from droblo.postgreSQL_poller.drobloPostgreSQLPoller import DrobloPostgreSQLPoller
from droblo.ORM.droblo_ORM import *
from droblo.filesystem.droblo_fs import get_recursive_file_list, get_file_dict, str_dict_to_path_dict, \
    get_connection_string, logon_start_install, logon_start_install_src_dst


# TODO: decide and implement how to handle paths data watch folders being moved, removed or renamed

class RestartWatcher(BaseException):
    """
    This exception is used to restart the watcher.
    """

    def __init__(self):
        super().__init__()


def recursively_str(r: bool):
    return " recursively" if r else ""


class DrobloWatcher:
    """
    Droblo Watcher class
    """
    observer = None
    no_signal = None

    def __init__(self, main_args):
        self.args = main_args
        self.host = socket.gethostname()
        self.session = get_session_from_args_or_conf(self.args)
        self.sqlite_session = get_sqlite_session()

        if self.session:
            self.paths = get_paths_from_args_or_session(self.args, self.session)
            self.excluded_paths = get_excluded_from_args_or_session(self.args, self.session)
            save_paths_to_db(self.session, self.args, self.paths, self.excluded_paths)
        else:
            self.paths = get_paths_from_args_or_session(self.args, self.sqlite_session)
            self.excluded_paths = get_excluded_from_args_or_session(self.args, self.sqlite_session)
        save_paths_to_db(self.sqlite_session, self.args, self.paths, self.excluded_paths)

        self.paths = str_dict_to_path_dict(self.paths)
        self.excluded_paths = str_dict_to_path_dict(self.excluded_paths)

        signal.signal(signal.SIGUSR2, self._reload)
        self.poller = None
        self.restart_watch = False
        self.status_dict = None

    def _end_watch(self):
        self.restart_watch = True
        raise RestartWatcher

    def _reload(self, signum, _):
        """
        restart the watchdog
        """
        logging.info("Reloading config (signal %s received)" % signum)
        postgre_session = get_session_from_args_or_conf(self.args)
        upsert_sqlite_data_to_postgresql(host=self.host, sqlite_session=self.sqlite_session,
                                         postgre_session=postgre_session)
        self._end_watch()

    def _start_poller(self):
        logging.info("Starting the poller")
        do_poll = threading.Event()
        connection_string = get_connection_string(self.args, obfuscated=True)
        self.poller = DrobloPostgreSQLPoller(connection_string=connection_string,
                                             droblo_poll=do_poll, interval=self.args.interval)
        do_poll.set()
        self.poller.start()
        return do_poll

    def watch(self):
        """
        Starts the watchdog. All files info are updated in DB upon FS event.
        """

        if DrobloStatus(self.sqlite_session).is_watcher_running(self.host):
            logging.critical("The watcher is already running on %s." % self.host)
            sys.exit(1)

        do_poll = self._start_poller()
        self.no_signal = True
        event_handler = PostgresEventHandler(self.session, self.sqlite_session, self.paths, self.excluded_paths)
        self.observer = Observer()
        exception_raised = False

        for path, recursive in self.paths.items():
            if path.exists():
                logging.info("Adding %s to the list of%s watched folders." % (path, recursively_str(r=recursive)))
                for xp in self.excluded_paths:
                    if path in xp.parents:
                        logging.info("\tEvents triggered under %s are ignored%s." % (xp, recursively_str(r=recursive)))
                self.observer.schedule(event_handler, str(path), recursive=recursive)
            else:
                logging.warning("Skipping %s : path does not exists." % path)

        try:
            self.observer.start()
        except OSError:
            logging.error("Permission denied on one of the given paths. Exiting.")
            sys.exit(-1)

        try:
            self._initialize_status_table()
        except DrobloBroadException:
            logging.critical("Unable to initialize the status table")
            sys.exit(-1)

        try:
            status_watchdog = 0
            while self.no_signal:
                if status_watchdog == 20:
                    self._heartbeat_status()
                    status_watchdog = 0
                time.sleep(0.5)
                status_watchdog += 1
        except RestartWatcher:
            logging.info("Stopping the observer (restart request)")
            if self.sqlite_session.info.get('has_flushed', False):
                self.sqlite_session.commit()
            if self.session.info.get('has_flushed', False):
                self.session.rollback()
            self.observer.stop()
        except KeyboardInterrupt:
            logging.info("Stopping the observer (KeyboardInterrupt)")
            self.observer.stop()
        except DrobloBroadException:
            logging.exception("Unknown exception")
            exception_raised = True
        finally:
            logging.info("Exiting watch %s" % (" in error" if exception_raised else ""))
            self.observer.join()
            logging.debug("Observer joined")
            do_poll.clear()
            if not self.restart_watch:
                self._exit_droblod(exception_raised)
            else:
                self.restart_watch = False
                self.session = get_session_from_args_or_conf(self.args)
                self.watch()

    def _exit_droblod(self, exception_raised):
        logging.info("Exiting droblod.")
        self.status_dict["end_time"] = datetime.now()
        DrobloStatus(self.session, self.status_dict)
        DrobloStatus(self.sqlite_session, self.status_dict)
        logging.debug("Status table cleaned.")
        sys.exit(-1 if exception_raised else 0)

    def _heartbeat_status(self):
        self.status_dict["update_time"] = datetime.now()
        DrobloStatus(self.session, self.status_dict)
        DrobloStatus(self.sqlite_session, self.status_dict)
        logging.debug("Heartbeat sent on Status table.")

    def _initialize_status_table(self):
        self.status_dict = {"host": self.host,
                            "status": "watch",
                            "start_time": datetime.now(),
                            "update_time": datetime.now(),
                            "end_time": None
                            }
        DrobloStatus(self.session, self.status_dict)
        DrobloStatus(self.sqlite_session, self.status_dict)
        logging.debug("Status table initialized.")

    def show_config(self):
        """
        Displays information about the current setup on the command line
        """
        print()
        print("The database connection string is :")
        print("\t%s" % get_connection_string(self.args, obfuscated=True))
        print()
        if len(self.paths):
            print("The following paths are being watched:")
            for path, recursive in self.paths.items():
                print("\t%s%s" % (path, recursively_str(r=recursive)))
        else:
            print("No paths are being watched")
        print()
        if len(self.excluded_paths):
            print("The following paths are excluded:")
            for path, recursive in self.excluded_paths.items():
                print("\t%s%s" % (path, recursively_str(r=recursive)))
        else:
            print("No paths are excluded")

    def populate(self):
        """
        recursively parse all files in all paths, removing excluded paths from DB.
        :return:
        """
        if not self.session:
            logging.critical("Populate mode requires connectivity to the PostgreSQL server.")
            sys.exit(-1)

        status_dict = {"host": self.host,
                       "status": "populate",
                       "start_time": datetime.now(),
                       "update_time": None,
                       "end_time": None
                       }

        DrobloStatus(self.session, status_dict)
        if not DrobloStatus(self.session).is_clean(self.host):
            logging.warning("The watcher is not running.")
            logging.warning("It is highly recommended to populate while it is active.")

        file_count = 0
        for path, recursive in self.paths.items():
            if recursive:
                fl = get_recursive_file_list(path, self.excluded_paths)
            else:
                fl = [x for x in Path(path).iterdir() if not x.is_dir()]
            for file in fl:
                logging.info("Populating : %s" % file)
                try:
                    data = get_file_dict(str(file))
                    data["host"] = self.host
                    Files(self.session, data)
                    file_count += 1
                except FileNotFoundError:
                    pass
                except DrobloBroadException:
                    logging.exception("Other exception. To investigate")
        for excluded in self.excluded_paths.keys():
            Files(self.session, {"host": self.host, "deleted": excluded})
        status_dict["end_time"] = datetime.now()
        DrobloStatus(self.session, status_dict)
        return file_count


def _logon_start_install_help():
    src, install_path = logon_start_install_src_dst(root=Path(__file__).parent)
    if sys.platform == 'linux':
        return f"Installs systemd unit file {src} to {install_path}"
    elif sys.platform == 'darwin':
        return f"Installs plaunchd plist file {src} to {install_path}"
    elif sys.platform == 'win32':
        return f"Registers scheduled task {src} to the windows task scheduler."
    else:
        return f"logon_start for {sys.platform} is not implemented at the moment."


def parse_arguments():
    """
    parse cli arguments
    :return:
    """
    parser = argparse.ArgumentParser(description='File system watcher inserting files info to DB.',
                                     formatter_class=argparse.ArgumentDefaultsHelpFormatter)
    parser.add_argument("-H", "--db_host", help="Database server hostname", default="localhost", type=str)
    parser.add_argument("-P", "--db_port", help="Database server port", default="5432", type=int)
    parser.add_argument("-N", "--db_name", help="Database name", default="droblo", type=str)
    parser.add_argument("-U", "--db_username", help="Database username", default="droblo", type=str)
    parser.add_argument("-W", "--db_password", help="Database password", type=str)
    parser.add_argument("-i", "--interval", help="Database server polling interval in seconds", type=int, default=3)

    parser.add_argument("-p", "--paths",
                        help="Paths to watch. If none is given, will use the values in table paths.\n"
                             "One can specify the path recursivity by suffixing it with ',1'.\n"
                             "Recursivity is off by default for watch paths.", nargs='+', type=str)
    parser.add_argument("-x", "--excluded",
                        help="Paths to exclude from the watch. If none is given, will use the values in table paths.\n"
                             "One can specify the path non-recursivity by suffixing it with ',0'\n"
                             "Recursivity is on by default for exclusion paths.", nargs='+', type=str)
    parser.add_argument("-s", "--save", help="Insert or update given paths in DB.", action="store_true")
    parser.add_argument("-S", "--purge_and_save", help="Purge then insert given paths in DB.", action="store_true")
    parser.add_argument("--show_config", help="Displays the current configuration.", action="store_true")
    parser.add_argument("--populate", help="Populate the database in non-watching mode.", action="store_true")
    parser.add_argument("--logon_start", help=_logon_start_install_help(), action="store_true")
    parser.add_argument("-v", "--verbose", help="Verbose (logs events information).", action="store_true")
    parser.add_argument("-d", "--debug", help="Debugging (logs debugging information).", action="store_true")
    parser.add_argument("-l", "--log_file", help="Path to log file.", type=Path)
    parser.add_argument("-sh", "--syslog_host", help="Syslog hostname.", type=str)
    parser.add_argument("-sp", "--syslog_port", help="Syslog port.", type=int, default=514)
    return parser.parse_args()


def main():
    """
    Droblo entry point
    """
    args = parse_arguments()
    try:
        pap = PaPLogger(verbose_fmt=True, logfile_with_hostname=True)
        pap.level = DEBUG if args.debug else INFO
        pap.verbose_fmt = args.verbose
        pap.log_file = args.log_file
        pap.syslog_port = args.syslog_port
        pap.syslog_host = args.syslog_host

        if args.logon_start:
            logon_start_install(root=Path(__file__).parent)

        if args.populate:
            logging.info("Starting to populate")
            start_time = datetime.now()
            files_count = DrobloWatcher(args).populate()
            end_time = datetime.now()
            logging.info("Checked %d files in %s" % (files_count, (end_time - start_time)))
        elif args.show_config:
            DrobloWatcher(args).show_config()
        else:
            logging.info("Starting to watch")
            while True:
                DrobloWatcher(args).watch()
                logging.info("Force restarting the watcher. Pretty bad")
    except KeyboardInterrupt:
        logging.info("Exiting (KeyboardInterrupt)")
        sys.exit(0)
    except DrobloBroadException:
        logging.exception("Other exception. To investigate")
        logging.shutdown()
        sys.exit(-1)


if __name__ == "__main__":
    main()



# coding=utf-8
"""
Filesystem functions.
"""
import logging
import json
import re
import sys
from datetime import datetime
from pathlib import Path
import hashlib


def _hash_bytestr_iter(bytes_iter: bytes, hashing_function, as_hex_str=False):
    for block in bytes_iter:
        hashing_function.update(block)
    return hashing_function.hexdigest() if as_hex_str else hashing_function.digest()


def _file_as_blockiter(file, block_size=65536) -> bytes:
    with file:
        block = file.read(block_size)
        while len(block) > 0:
            yield block
            block = file.read(block_size)


def get_file_dict(path: (Path, str)) -> dict:
    """
    return hash, fs time and file size for a given file.
    :param path:
    :return:
    """
    p = Path(path) if isinstance(path, str) else path
    if p.exists():
        atime = datetime.fromtimestamp(p.stat().st_atime)
        ctime = datetime.fromtimestamp(p.stat().st_ctime)
        mtime = datetime.fromtimestamp(p.stat().st_mtime)
        size = p.stat().st_size
        file_hash = _hash_bytestr_iter(_file_as_blockiter(p.open(mode="rb")), hashlib.sha256())
        return {"file": path, "hash": file_hash,
                "atime": atime, "ctime": ctime, "mtime": mtime, "size": size
                }
    else:
        raise FileNotFoundError("%s does not exist." % path)


def is_excluded(path: (Path, str), excluded_paths) -> bool:
    """
    returns a True if path is in exclusion list, taking recursivity into account.
    :param path:
    :param excluded_paths:
    :return:
    """
    path = str(path) if isinstance(path, Path) else path
    for e, r in excluded_paths.items():
        if (path.startswith(str(e)) and r) or path == e:
            return True
    return False


def get_recursive_file_list(path: (Path, str), excluded_paths) -> list:
    """
    returns a list of all files in path and subdirectory, minus excluded paths.
    :param path:
    :param excluded_paths:
    :return:
    """
    fl = []
    p = Path(path) if isinstance(path, str) else path
    for i in p.glob('**/*'):
        if i.is_file() and not is_excluded(str(i), excluded_paths):
            fl.append(i)
    return fl


def get_config_path() -> Path:
    """
    get config path , if possible by using XDG spec (?)
    :return: pathlib's Path object
    """
    droblo_config_folder = Path("~") / ".droblo"
    droblo_config_folder = droblo_config_folder.expanduser()
    if not droblo_config_folder.exists():
        droblo_config_folder.mkdir()
        logging.info(f"Created {droblo_config_folder}")
    return droblo_config_folder


def dump_connection_string(conn_string: str) -> bool:
    """
    dump connection string!
    :param conn_string:
    """

    try:
        (get_config_path() / "droblo.json").write_text(json.dumps(conn_string))
        return True
    except IOError:
        logging.exception("Failed to save connection string")
    return False


def load_connection_string() -> [str, None]:
    """
    load connection string!
    """

    try:
        conn_string_config = get_config_path() / "droblo.json"
        if conn_string_config.exists():
            with conn_string_config.open() as droblo_json:
                return json.load(droblo_json)
    except IOError:
        logging.error("Failed to load connection string")
    return None


def same_parent(src: (Path, str), dst: (Path, str)):
    """
    Test 2 paths and check if they have the same parent
    :param src:
    :param dst:
    :return:
    """
    src = Path(src) if isinstance(src, str) else src
    dst = Path(dst) if isinstance(dst, str) else dst
    return src.parent == dst.parent


def str_dict_to_path_dict(orm_dict: dict):
    """
    converts orm_dict keys from string to Path
    :param orm_dict:
    :return:
    """
    paths_dict = {}
    for p, r in orm_dict.items():
        paths_dict[Path(p).expanduser()] = r
    return paths_dict


def get_connection_string(main_args, obfuscated=False):
    """
    returns postgre connection string.
    :param obfuscated:
    :param main_args:
    :return:
    """
    conn_string = load_connection_string()
    if conn_string:
        if obfuscated:
            return re.sub(r'(postgresql://.*:).*(@.*)', r"\1password_is_obfucasted\2", conn_string)
        else:
            return conn_string
    elif main_args.db_password:
        if obfuscated:
            return "postgresql://%s:no_pwd_in_logs@%s:%s/%s" % (main_args.db_username,
                                                                main_args.db_host,
                                                                main_args.db_port,
                                                                main_args.db_name)
        else:
            return "postgresql://%s:%s@%s:%s/%s" % (main_args.db_username,
                                                    main_args.db_password,
                                                    main_args.db_host,
                                                    main_args.db_port,
                                                    main_args.db_name)
    else:
        logging.error("No password provided on the command line and config file not found.")
        sys.exit(1)


def logon_start_install_src_dst(root):
    """
    Returns src file and dst folder for logon startup setup, based on the OS.
    :return:
    """
    data = root / "data"
    if not data.exists():
        logging.critical("That's a bug.")
        sys.exit(-2)

    if sys.platform == 'linux':
        source_path = data / 'droblod.user.service'
        install_path = Path("~/.config/systemd/user/").expanduser()
    elif sys.platform == 'darwin':
        source_path = data / 'droblod.plist'
        install_path = Path("~/Library/LaunchAgents/").expanduser()
    elif sys.platform == 'win32':
        source_path = data / 'Droblod.xml'
        install_path = Path(r"C:\temp")
    else:
        source_path = None
        install_path = None
    return source_path, install_path


def logon_start_install(root):
    """
    Installs the logon start file (and creates the scheduled task under Windows.)
    """
    logging.info("Setting up logon start.")
    src, dst = logon_start_install_src_dst(root=root)
    if src and dst:
        if not dst.exists():
            logging.warning(f"Creating {dst}")
            dst.mkdir(parents=True)
        dst_file = dst / src.name
        if not dst_file.exists():
            dst_file.write_text(src.read_text())
            logging.info(f"{src} saved to {dst}")
            if sys.platform != 'darwin':
                import subprocess
                if sys.platform == "linux":
                    _cmd = ["systemctl", "--user", "--enable", "droblod.user"]
                else:
                    _cmd = ["schtasks.exe", "/create", "/XML", str(dst_file), "/tn", "droblod"]

                rtn = subprocess.run(_cmd, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
                if rtn.returncode == 0:
                    logging.info("Logon start successfully setup")
                else:
                    logging.warning("logon start setup failed")
                    logging.warning(rtn.stdout.decode())
                    logging.warning(rtn.stderr.decode())
        else:
            logging.warning(f"{dst_file} already exists.")
    else:
        logging.warning(f"logon_start for {sys.platform} is not implemented at the moment.")



# coding=utf-8
"""
Reporting queries
"""
from datetime import datetime, timedelta

from sqlalchemy import func, or_
import pandas as pd

from droblo.ORM.droblo_ORM import Files, DrobloStatus


class DrobloReports:
    """
    Reporting queries returned as pandas DataFrame
    """

    def __init__(self, session):
        self.session = session
        self.con = self.session.bind.engine

    def get_list_hosts(self, as_list=False) -> [list, pd.DataFrame]:
        """

        :param as_list:
        :return:
        """
        q = self.session.query(func.distinct(Files.host).label("Hosts"))
        if as_list:
            return [h.Hosts for h in q.all()]
        else:
            return pd.read_sql(sql=q.statement, con=self.con)

    def get_duplicated_files(self, host=None, min_size=None, modified_after=None, order=True):
        """
        Returns the list of duplicated files
        :param host:
        :param min_size:
        :param modified_after:
        :return:
        """
        dup = func.count(Files.hash).label("dup")
        sub_q = self.session.query(dup, Files.hash.label("hash")). \
            group_by(Files.hash). \
            having(dup > 1).subquery()

        q = self.session.query(Files, sub_q.c.dup). \
            join(sub_q, Files.hash == sub_q.c.hash)
        if host:
            q = q.filter(Files.host == host)
        if min_size:
            q = q.filter(Files.size > min_size)
        if modified_after:
            q = q.filter(Files.mtime > modified_after)
        if order:
            q = q. \
                order_by(sub_q.c.dup.desc()). \
                order_by(Files.size.desc()). \
                order_by(Files.ctime.desc())  # we want to highlight the most recent duplication.

        return pd.read_sql(sql=q.statement, con=self.con)

    def get_potential_disk_space_savings(self, host=None):
        """
        :return
        :param host:
        :return:
        """
        df = self.get_duplicated_files(host=host)
        df = pd.DataFrame(df.groupby(by=["host"])["size"].sum())
        if host:
            return df.query("host=='%s'" % host)
        else:
            return df

    def get_inactive_watchers(self) -> pd.DataFrame:
        """

        :return:
        """
        update_time_threshold = datetime.now() + timedelta(seconds=-60)
        q = self.session.query(DrobloStatus.host,
                               DrobloStatus.start_time,
                               DrobloStatus.update_time,
                               DrobloStatus.end_time). \
            filter(DrobloStatus.status == "watch"). \
            filter(or_(~DrobloStatus.end_time.is_(None), DrobloStatus.update_time < update_time_threshold))
        return pd.read_sql(sql=q.statement, con=self.con)



# coding=utf-8
"""
Report generation
"""
import argparse
import socket
import sys
import logging
from pathlib import Path
from collections import OrderedDict
from pap_logger import *
from sqlalchemy.exc import OperationalError

from droblo.ORM.droblo_ORM import get_session, DrobloBroadException
from droblo.filesystem.droblo_fs import load_connection_string
from droblo.ORM.droblo_reports_ORM import DrobloReports

__REPORT_EXT__ = ["csv"]


def parse_arguments():
    """
    parse cli arguments
    :return:
    """
    parser = argparse.ArgumentParser(description='File system watcher inserting files info to DB.',
                                     formatter_class=argparse.ArgumentDefaultsHelpFormatter)
    parser.add_argument("-a", "--all", help="Report on all hosts.", action="store_true")
    parser.add_argument("-A", "--ALL", help="Report on all hosts, one file per host.", action="store_true")
    parser.add_argument("-r", "--reports_path", help="Path where to store reports.", type=Path)
    parser.add_argument("-t", "--reports_type", help="Reports type.", choices=__REPORT_EXT__, default="csv")

    parser.add_argument("-v", "--verbose", help="Verbose.", action="store_true")
    parser.add_argument("-d", "--debug", help="Debugging.", action="store_true")
    parser.add_argument("-l", "--log_file", help="Path to log file.", type=Path)
    parser.add_argument("-sh", "--syslog_host", help="Syslog hostname.", type=str)
    parser.add_argument("-sp", "--syslog_port", help="Syslog port.", type=int, default=514)
    return parser.parse_args()


def main():
    """
    Droblo_reports entry point
    """
    args = parse_arguments()
    try:
        pap = PaPLogger(verbose_fmt=True, logfile_with_hostname=True)
        pap.level = DEBUG if args.debug else INFO
        pap.verbose_fmt = args.verbose
        pap.log_file = args.log_file
        pap.syslog_port = args.syslog_port
        pap.syslog_host = args.syslog_host
        try:
            session = get_session(connection_string=load_connection_string(), echo=args.debug)
            split_by_host = True if args.ALL else False
            reports = Reports(session=session, all=args.all)
            reports.get_results()
            reports.generate_reports(extension=args.reports_type,
                                     split_by_host=split_by_host,
                                     reports_path=args.reports_path)
        except OperationalError:
            logging.error(msg="Connection to database failed.")
            sys.exit(-1)
    except KeyboardInterrupt:
        logging.info("Exiting")
        sys.exit(0)
    except DrobloBroadException:
        logging.exception("Other exception. To investigate")
        logging.shutdown()
        sys.exit(-1)


class Reports:
    """
    Get data and generates reports
    """

    def __init__(self, session, all: bool):
        self.dr = DrobloReports(session=session)
        self.all = all
        self.host = socket.gethostname()
        self.reports = OrderedDict()
        self.hosts = self.dr.get_list_hosts(as_list=True) if self.all else []

    def get_results(self):
        """
        Get pandas DataFrame
        """
        if self.all:
            self.reports["Hosts"] = self.dr.get_list_hosts()
            self.reports["Hosts with inactive watchers"] = self.dr.get_inactive_watchers()
        dup = self.dr.get_duplicated_files() if self.all else self.dr.get_duplicated_files(host=self.host)
        self.reports["Duplicated files"] = dup
        ds = self.dr.get_potential_disk_space_savings() if self.all else self.dr.get_potential_disk_space_savings(host=self.host)
        self.reports["Potential disk space saving"] = ds

    def generate_reports(self, extension: str, reports_path: Path, split_by_host: bool):
        for name, df in self.reports.items():
            print(name)
            print()
            print(df)
            print()
            print()


if __name__ == "__main__":
    main()



# coding=utf-8
"""
This modules handles everything DB related for this project
"""
import logging
import socket
import sys

from sqlalchemy import create_engine
from sqlalchemy import Column, Index
from sqlalchemy import Integer, String, DateTime, Boolean, BigInteger, BLOB
from sqlalchemy import update, func, delete, insert

from sqlalchemy.orm import sessionmaker, scoped_session
from sqlalchemy.dialects.postgresql import insert as pginsert
from sqlalchemy.ext.declarative import declarative_base
from sqlalchemy.sql.elements import True_, False_
from sqlalchemy.exc import OperationalError, ResourceClosedError

from watchdog.events import FileSystemEventHandler, FileCreatedEvent

from droblo.filesystem.droblo_fs import get_file_dict, get_recursive_file_list, is_excluded, get_config_path, \
    get_connection_string, dump_connection_string

Base = declarative_base()

__SECONDS_THRESHOLD_FOR_CLEAN__ = 11
__PSYCOPG2__ = 'psycopg2'


class DrobloBroadException(BaseException):
    """
    BaseException override for pep8 compliance.
    """

    def __init__(self):
        super().__init__()


class DrobloConnectivityLostException(DrobloBroadException):
    """
        BaseException override for pep8 compliance.
        """

    def __init__(self):
        super().__init__()


def get_session_driver_name(session):
    return session.bind.dialect.driver


def get_session(connection_string: str, echo=False, pool_pre_ping=True):
    """
    return SQLAlchemy session object.
    :param pool_pre_ping:
    :param connection_string:
    :param echo:
    :return:
    """
    engine = create_engine(connection_string, echo=echo, pool_pre_ping=pool_pre_ping)
    session_obj = scoped_session(sessionmaker(bind=engine))
    Base.metadata.create_all(engine)
    session = session_obj()
    session.connection()
    return session


def get_sqlite_session(echo=False):
    droblo_db = get_config_path() / 'droblo.db'

    return get_session(connection_string=f"sqlite:///{droblo_db}", echo=echo, pool_pre_ping=False)


def get_session_from_args_or_conf(main_args):
    """
    watch and populate common values handling.
    :param main_args:
    :return:
    """
    conn_str = get_connection_string(main_args)
    try:
        session = get_session(connection_string=conn_str, echo=main_args.debug)
    except OperationalError:
        obfuscated_conn_str = get_connection_string(main_args, obfuscated=True)
        logging.error(msg="Connection to database failed. Connection string was %s" % obfuscated_conn_str)
        logging.warning(msg="Falling back to sqlite only.")
        session = None
    else:
        dump_connection_string(conn_str)

    return session


def files_to_dict(f, deleted_files=False):
    if deleted_files:
        return {"deleted": f.file, "host": f.host}
    return {"file": f.file, "hash": f.hash, "host": f.host, "size": f.size, "mtime": f.mtime, "atime": f.atime,
            "ctime": f.ctime}


def get_excluded_from_args_or_session(main_args=None, session=None):
    if not main_args or not main_args.excluded:
        excluded_paths = Paths().get_excluded_paths(session)
    else:
        excluded_paths = {}
        for p in main_args.excluded:
            p_split = p.split(",")
            if len(p_split) > 1:
                excluded_paths[p_split[0]] = bool(p_split[1])
            else:
                excluded_paths[p] = True
    return excluded_paths


def get_paths_from_args_or_session(main_args=None, session=None):
    if not main_args or not main_args.paths:
        paths = Paths().get_watch_paths(session)
    else:
        paths = {}
        for p in main_args.paths:
            p_split = p.split(",")
            if len(p_split) > 1:
                paths[p_split[0]] = bool(p_split[1])
            else:
                paths[p] = False
    if not paths:
        logging.error("no paths configured; specify then on the command line")
        sys.exit(1)

    return paths


def upsert_sqlite_data_to_postgresql(host, sqlite_session, postgre_session):
    try:
        sqlite_paths = get_paths_from_args_or_session(session=sqlite_session)
        sqlite_excluded = get_excluded_from_args_or_session(session=sqlite_session)
        logging.info("Updating PostgreSQL paths table using sqlite paths table.")
        Paths().save_paths(postgre_session, paths=sqlite_paths, excluded=sqlite_excluded, purge=True)
    except OperationalError:
        logging.exception("Ouch OperationalError")
    except DrobloBroadException:
        logging.exception("Ouch DrobloBroadException")

    try:
        logging.info("Updating files table for new and updated files.")
        max_mod_time = postgre_session.query(func.max(Files.mtime)).filter(Files.host == host).scalar()
        sqlite_files_q = sqlite_session.query(Files).filter(Files.host == host)
        if max_mod_time:
            logging.info(f"The last modification time in PostgreSQL is {max_mod_time}")
            sqlite_files_q = sqlite_files_q.filter(Files.host >= max_mod_time)
        files_data = sqlite_files_q.all()
        logging.info(f"Inserting {len(files_data)} entries from sqlite to PostgreSQL.")
        Files(postgre_session, [files_to_dict(f) for f in files_data])
        logging.info(f"Cleaning up sqlite files table.")
        sqlite_session.execute(delete(Files))
    except (OperationalError, ResourceClosedError):
        logging.exception("Ouch (OperationalError,ResourceClosedError). Rolling back.")
        sqlite_session.rollback()
    except DrobloBroadException:
        logging.exception("Ouch DrobloBroadException. Rolling back.")
        sqlite_session.rollback()

    try:
        logging.info("Updating files table for deleted files")
        deleted_files_data = sqlite_session.query(DeletedFiles).filter(DeletedFiles.host == host).all()
        logging.info(f"Deleting {len(deleted_files_data)} entries from sqlite to PostgreSQL.")
        Files(postgre_session).delete_from_list([files_to_dict(f, True) for f in deleted_files_data])
        logging.info(f"Cleaning up sqlite deleted_files table.")
        sqlite_session.execute(delete(DeletedFiles))
    except (OperationalError, ResourceClosedError):
        logging.exception("Ouch (OperationalError,ResourceClosedError). Rolling back.")
        sqlite_session.rollback()
    except DrobloBroadException:
        logging.exception("Ouch DrobloBroadException. Rolling back.")
        sqlite_session.rollback()


def save_paths_to_db(session, main_args, paths, excluded_paths):
    # if not session or session.bind.engine.connection.closed:
    if not session:
        logging.warning("No session or session is disconnected")
        return
    if main_args.save or main_args.purge_and_save:
        purge = True if main_args.purge_and_save else False
        Paths().save_paths(session, paths, excluded_paths, purge)
        logging.info("New paths configuration saved.")


class PostgresEventHandler(FileSystemEventHandler):
    """
    Implemented on top of watchdog FileSystemEventHandler, handles captures events to DB (via Files and Paths classes).
    """

    def __init__(self, session, sqlite_session, paths, excluded_paths):

        self.session = session
        self.sqlite_session = sqlite_session
        self.host = socket.gethostname()
        self.paths = paths
        self.excluded = excluded_paths

    @staticmethod
    def _get_what(event) -> str:
        return 'directory' if event.is_directory else 'file'

    def _send_recursive_fake_created_events(self, path):
        """
        Force created events (used when destination of on_moved is not excluded.)
        :param path:
        """
        for file in get_recursive_file_list(path, self.excluded):
            FileCreatedEvent(file)

    def on_moved(self, event):
        """
        watchdog file moved event handling.
        :param event:
        :return:
        """
        if is_excluded(event.dest_path, self.excluded):
            # in case the exclusion list was updated
            if self.session:
                try:
                    Files(self.session, {"host": self.host, "deleted": event.src_path})
                except DrobloConnectivityLostException:
                    # self.session.remove()
                    self.session = None
            Files(self.sqlite_session, {"host": self.host, "deleted": event.src_path})
            logging.debug("Deleting %s as its destination is in the exclusion list" % event.dest_path)
        elif is_excluded(event.src_path, self.excluded):
            logging.debug("Forcing created event %s as it is in the exclusion list" % event.src_path)
            if event.is_directory:
                self._send_recursive_fake_created_events(event.src_path)
            else:
                self.on_created(event)
        else:
            data = {"host": self.host, "src": event.src_path, "dst": event.dest_path}
            action = 'moved_dir' if event.is_directory else 'moved_file'
            data[action] = ""
            if self.session:
                try:
                    Files(self.session, data)
                except DrobloConnectivityLostException:
                    # self.session.remove()
                    self.session = None
            Files(self.sqlite_session, data)
            logging.info("Moved %s: from %s to %s", self._get_what(event), event.src_path, event.dest_path)

        if event.src_path in self.paths.keys():
            if is_excluded(event.dest_path, self.excluded):
                logging.debug("Removing %s from the config as it was moved to an excluded path" % event.src_path)
                if self.session:
                    Paths().delete_path(self.session, path=event.src_path)
                Paths().delete_path(self.sqlite_session, path=event.src_path)
            else:
                logging.info("Updating %s in the config" % event.src_path)
                if self.session:
                    Paths().update_path(self.session, src=event.src_path, dst=event.dest_path, excluded=False)
                Paths().update_path(self.sqlite_session, src=event.src_path, dst=event.dest_path, excluded=False)

        if event.src_path in self.excluded.keys():
            logging.debug("Updating %s in the config" % event.src_path)
            if self.session:
                Paths().update_path(self.session, src=event.src_path, dst=event.dest_path, excluded=True)
            Paths().update_path(self.sqlite_session, src=event.src_path, dst=event.dest_path, excluded=True)

    def on_created(self, event):
        """
        watchdog file created event handling
        :param event:
        :return:
        """
        what = self._get_what(event)

        if is_excluded(event.src_path, self.excluded):
            # in case the exclusion list was updated
            if self.session:
                try:
                    Files(self.session, {"host": self.host, "deleted": event.src_path})
                except DrobloConnectivityLostException:
                    # self.session.remove()
                    self.session = None
            Files(self.sqlite_session, {"host": self.host, "deleted": event.src_path})
            logging.debug("Deleting %s as it is in the exclusion list" % event.src_path)
            return

        if what == 'file':
            try:
                data = get_file_dict(event.src_path)
                data["host"] = self.host
                if self.session:
                    try:
                        Files(self.session, data)
                    except DrobloConnectivityLostException:
                        # self.session.remove()
                        self.session = None
                Files(self.sqlite_session, data)
            except FileNotFoundError:
                return
        logging.info("Created %s: %s", what, event.src_path)

    def on_deleted(self, event):
        """
        watchdog file deleted event handling
        :param event:
        """
        if self.session:
            try:
                Files(self.session, {"host": self.host, "deleted": event.src_path})
            except DrobloConnectivityLostException:
                # # self.session.remove()
                self.session = None
        Files(self.sqlite_session, {"host": self.host, "deleted": event.src_path})
        if is_excluded(event.src_path, self.excluded):
            logging.debug("Not logging %s as it is in the exclusion list" % event.src_path)
        else:
            logging.info("Deleted %s: %s", self._get_what(event), event.src_path)

        if event.src_path in self.paths.keys():
            logging.info("Deleting %s from the config as well" % event.src_path)
            if self.session:
                Paths().delete_path(self.session, path=event.src_path)
            Paths().delete_path(self.sqlite_session, path=event.src_path)

        if event.src_path in self.excluded.keys():
            logging.debug("Leaving %s in the config as we want to prevent it being watched if is later re-created")

    def on_modified(self, event):
        """
        watchdog file modified event handling
        :param event:
        :return:
        """
        what = self._get_what(event)

        if is_excluded(event.src_path, self.excluded):
            # in case the exclusion list was updated
            if self.session:
                try:
                    Files(self.session, {"host": self.host, "deleted": event.src_path})
                except DrobloConnectivityLostException:
                    # self.session.remove()
                    self.session = None
            Files(self.sqlite_session, {"host": self.host, "deleted": event.src_path})
            logging.debug("Deleting %s as it is in the exclusion list" % event.src_path)
        elif what == 'file':
            try:
                data = get_file_dict(event.src_path)
                data["host"] = self.host
                if self.session:
                    try:
                        Files(self.session, data)
                    except DrobloConnectivityLostException:
                        # self.session.remove()
                        self.session = None
                Files(self.sqlite_session, data)
                logging.info("Modified %s: %s", what, event.src_path)
            except FileNotFoundError:
                return
        else:
            # We don't store info on directories
            pass

        # if event.src_path in self.paths.keys():
        #     src = Path(event.src_path)
        #     dst = Path(event.dest_path)
        #     if src.parent == dst.parent:
        #         logging.info("Updating %s in the config" % event.src_path)
        #         Paths().update_paths(self.session, src=event.src_path, dst=event.dest_path, excluded=False)
        #
        # if event.src_path in self.excluded.keys():
        #     src = Path(event.src_path)
        #     dst = Path(event.dest_path)
        #     if src.parent == dst.parent:
        #         logging.debug("Updating %s in the config" % event.src_path)
        #         Paths().update_paths(self.session, src=event.src_path, dst=event.dest_path, excluded=True)


class Paths(Base):
    """
    Paths
    Table where we store the different hosts configuration regarding watch or exclusion paths along a recursivity flag.
    Not to be confounded with pathlib Path (which we make sure to not use in this file for this exact reason)
    """
    __tablename__ = 'paths'
    id = Column(Integer, primary_key=True, autoincrement=True)
    host = Column(String, nullable=False, unique=False)
    path = Column(String, nullable=False, unique=False)
    recursive = Column(Boolean, nullable=False, unique=False)
    excluded = Column(Boolean, nullable=False, unique=False)
    paths_index = Index("paths_index", host, path, unique=True)

    @staticmethod
    def get_watch_paths(session):
        """
        :param session:
        :return:
        """
        if not session:
            logging.warning("No session or session is disconnected")
            return

        return dict(session.query(Paths.path, Paths.recursive).
                    filter(Paths.host == socket.gethostname()).
                    filter(Paths.excluded == False_()).all())

    @staticmethod
    def get_excluded_paths(session):
        """

        :param session:
        :return:
        """
        if not session:
            logging.warning("No session or session is disconnected")
            return

        return dict(session.query(Paths.path, Paths.recursive).
                    filter(Paths.host == socket.gethostname()).
                    filter(Paths.excluded == True_()).all())

    @staticmethod
    def save_paths(session, paths, excluded, purge):
        """
        insert or update into status
        :param purge:
        :param excluded:
        :param paths:
        :param session:
        """

        if not session:
            logging.warning("No session or session is disconnected")
            return

        host = socket.gethostname()
        if purge:
            delete_stmt = delete(Paths).where(Paths.host == host)
            session.execute(delete_stmt)
        if get_session_driver_name(session) == __PSYCOPG2__:
            for path, recursive in paths.items():
                insert_stmt = pginsert(Paths).values(host=host, path=path, recursive=recursive, excluded=False)
                update_stmt = insert_stmt.on_conflict_do_update(
                    index_elements=Paths.paths_index.columns, set_=dict(recursive=recursive, excluded=False)
                )
                session.execute(update_stmt)

            for path, recursive in excluded.items():
                insert_stmt = pginsert(Paths).values(host=host, path=path, recursive=recursive, excluded=True)
                update_stmt = insert_stmt.on_conflict_do_update(
                    index_elements=Paths.paths_index.columns, set_=dict(recursive=recursive, excluded=True)
                )
                session.execute(update_stmt)
        else:
            for path, recursive in paths.items():
                delete_stmt = delete(Paths).where(Paths.host == host). \
                    where(Paths.path == path).where(Paths.excluded.is_(False))
                insert_stmt = insert(Paths).values(host=host, path=path, recursive=recursive, excluded=False)

                session.execute(delete_stmt)
                session.execute(insert_stmt)

            for path, recursive in excluded.items():
                delete_stmt = delete(Paths).where(Paths.host == host). \
                    where(Paths.path == path).where(Paths.excluded.is_(True))
                insert_stmt = insert(Paths).values(host=host, path=path, recursive=recursive, excluded=True)

                session.execute(delete_stmt)
                session.execute(insert_stmt)

        try:
            session.commit()
        except (OperationalError, ResourceClosedError):
            logging.exception("Rolling back save_paths ((OperationalError,ResourceClosedError))")
            session.rollback()
        except DrobloBroadException:
            logging.exception("Rolling back save_paths (DrobloBroadException)")
            session.rollback()

    @staticmethod
    def update_path(session, src: str, dst: str, excluded: bool):
        """
        Update paths in the config (called following on_modifed event)
        :param session:
        :param src:
        :param dst:
        :param excluded:
        """
        if not session:
            logging.warning("No session or session is disconnected")
            return

        host = socket.gethostname()

        update_path_stmt = update(Paths).values(path=dst). \
            where(Paths.host == host).where(Paths.path == src).where(Paths.excluded is excluded)
        session.execute(update_path_stmt)

        try:
            session.commit()
        except (OperationalError, ResourceClosedError):
            logging.exception("Rolling back update_path ((OperationalError,ResourceClosedError))")
            session.rollback()
        except DrobloBroadException:
            logging.exception("Rolling back update_path (DrobloBroadException)")
            session.rollback()

    @staticmethod
    def delete_path(session, path: str):
        """
        Deletes paths from config (called following a on_deleted event)
        :param path:
        :param session:
        """
        if not session:
            logging.warning("No session or session is disconnected")
            return

        host = socket.gethostname()

        delete_path_stmt = delete(Paths). \
            where(Paths.host == host). \
            where(Paths.path == path). \
            where(Paths.excluded is False)
        session.execute(delete_path_stmt)

        try:
            session.commit()
        except (OperationalError, ResourceClosedError):
            logging.exception("Rolling back delete_path ((OperationalError,ResourceClosedError))")
            session.rollback()
        except DrobloBroadException:
            logging.exception("Rolling back delete_path (DrobloBroadException)")
            session.rollback()


class DrobloStatus(Base):
    """
    DrobloStatus
    Table to keep track of the different droblo processes.
    """
    __tablename__ = 'status'
    id = Column(Integer, primary_key=True, autoincrement=True)
    host = Column(String, nullable=False, unique=False)
    status = Column(String, nullable=False, unique=False)
    start_time = Column(DateTime, nullable=False, unique=False)
    update_time = Column(DateTime, nullable=True, unique=False)
    end_time = Column(DateTime, nullable=True, unique=False)
    status_index = Index("status_index", host, status, unique=True)

    def _pg_upsert(self, data):
        """
        insert or update into status
        :param data:
        """
        insert_stmt = pginsert(DrobloStatus).values(host=data["host"],
                                                    status=data["status"],
                                                    start_time=data["start_time"],
                                                    update_time=data["update_time"],
                                                    end_time=data["end_time"])
        do_update_stmt = insert_stmt.on_conflict_do_update(
            index_elements=DrobloStatus.status_index.columns,
            set_=dict(start_time=data["start_time"],
                      update_time=data["update_time"],
                      end_time=data["end_time"])
        )
        self.session.execute(do_update_stmt)
        self.session.commit()

    def _delete_and_insert(self, data):
        """
        insert or update into status
        :param data:
        """
        delete_stmt = delete(DrobloStatus). \
            where(DrobloStatus.host == data["host"]). \
            where(DrobloStatus.status == data["status"])

        insert_stmt = insert(DrobloStatus).values(host=data["host"],
                                                  status=data["status"],
                                                  start_time=data["start_time"],
                                                  update_time=data["update_time"],
                                                  end_time=data["end_time"])

        self.session.execute(delete_stmt)
        self.session.execute(insert_stmt)
        self.session.commit()

    def __init__(self, session, data=None):
        if session:
            self.session = session
        else:
            logging.warning("No session or session is disconnected")
            return
        if data:
            driver_name = get_session_driver_name(session=self.session)
            try:
                if driver_name == __PSYCOPG2__:
                    self._pg_upsert(data)
                else:
                    self._delete_and_insert(data)
            except (OperationalError, ResourceClosedError):
                logging.warning("Failed to update status on %s ((OperationalError,ResourceClosedError))" % driver_name)
                logging.warning("Rolling back")
                self.session.rollback()
            except DrobloConnectivityLostException:
                logging.warning("Failed to update status on %s (DrobloConnectivityLostException)" % driver_name)
                logging.warning("Rolling back")
                self.session.rollback()
            except DrobloBroadException:
                logging.warning("Failed to update status on %s (DrobloBroadException)" % driver_name)
                logging.warning("Rolling back")
                self.session.rollback()

    def is_clean(self, host):
        """
        Check if the watcher is already on host started before populating.
        If the watcher is not started since at least __SECONDS_THRESHOLD_FOR_CLEAN__ seconds, return false.
        This is a totally arbitrary decision. :P
        :param host:
        :return:
        """

        if self.session and self.is_watcher_running(host):
            watch_update_time = self.session.query(DrobloStatus.update_time). \
                filter(DrobloStatus.status == "watch"). \
                filter(DrobloStatus.host == host).scalar()
            populate_start_time = self.session.query(DrobloStatus.start_time). \
                filter(DrobloStatus.status == "populate"). \
                filter(DrobloStatus.host == host).scalar()
            if (populate_start_time - watch_update_time).seconds < __SECONDS_THRESHOLD_FOR_CLEAN__:
                return True
        return False

    def is_watcher_running(self, host):
        """
        Check if the watcher is running for host, by making sure the end time is not set
        :param host:
        :return:
        """

        watch_end_time = self.session.query(DrobloStatus.end_time). \
            filter(DrobloStatus.status == "watch"). \
            filter(DrobloStatus.host == host).all()

        if not len(watch_end_time):
            return False
        return False if watch_end_time else True


class DeletedFiles(Base):
    """
    files
    """
    __tablename__ = 'deleted_files'
    id = Column(Integer, primary_key=True, autoincrement=True)
    host = Column(String(32), nullable=False, unique=False)
    file = Column(String, nullable=False, unique=True)


class Files(Base):
    """
    files
    """
    __tablename__ = 'files'
    id = Column(Integer, primary_key=True, autoincrement=True)
    host = Column(String(32), nullable=False, unique=False)
    file = Column(String, nullable=False, unique=True)
    hash = Column(BLOB(32), nullable=False, unique=False)
    atime = Column(DateTime, nullable=False, unique=False)
    ctime = Column(DateTime, nullable=False, unique=False)
    mtime = Column(DateTime, nullable=False, unique=False)
    size = Column(BigInteger, nullable=False, unique=False)
    file_index = Index("file_index", host, file, unique=True)
    hash_index = Index("hash_index", host, hash, unique=False)
    size_index = Index("size_index", host, size, unique=False)

    def __init__(self, session, data=None):
        self.session = session
        self.driver_name = get_session_driver_name(self.session)
        if data and self.session:
            if isinstance(data, dict):
                self._handle_data(data)
            elif isinstance(data, list):
                for d in data:
                    self._handle_data(d)

    def _handle_data(self, data):
        if "hash" in data:  # on_created, on_modified
            self._upsert(data)
        elif "moved_dir" in data:  # on_moved
            if self.driver_name == __PSYCOPG2__:
                self._regex_update(data)
            else:
                self._sqlite_update(data)
        elif "moved_file" in data:  # on_moved
            # TODO : Check why we used to have below line

            # self._delete({"host": data["host"], "deleted": data["src"]})
            if self.driver_name == __PSYCOPG2__:
                self._regex_update(data)
            else:
                self._sqlite_update(data)
        else:  # on_deleted
            self._delete(data)

    def _upsert(self, data):
        if self.driver_name == __PSYCOPG2__:
            self._pg_upsert(data)
        else:
            self._delete_and_insert(data)

    def _pg_upsert(self, data):
        """
        insert or update into files
        :param data:
        """
        try:
            insert_stmt = pginsert(Files).values(host=data["host"],
                                                 file=data["file"], hash=data["hash"],
                                                 atime=data["atime"], ctime=data["ctime"],
                                                 mtime=data["mtime"], size=data["size"]
                                                 )
            do_update_stmt = insert_stmt.on_conflict_do_update(index_elements=Files.file_index.columns,
                                                               set_=dict(hash=data["hash"],
                                                                         atime=data["atime"], ctime=data["ctime"],
                                                                         mtime=data["mtime"], size=data["size"]
                                                                         )
                                                               )

            self.session.execute(do_update_stmt)
            self.session.commit()
        except (OperationalError, ResourceClosedError):
            logging.error("Failed to upsert on %s ((OperationalError,ResourceClosedError))" % self.driver_name)
            logging.error("Rolling back.")
            self.session.rollback()
            raise DrobloConnectivityLostException()
        except DrobloBroadException:
            logging.error("Failed to upsert on %s (DrobloBroadException)" % self.driver_name)
            logging.error("Rolling back.")
            self.session.rollback()
            raise DrobloConnectivityLostException()

    def _delete_and_insert(self, data):
        """
        insert or update into files
        :param data:
        """
        try:
            delete_stmt = delete(Files).where(Files.host == data["host"]).where(Files.file == data["file"])

            insert_stmt = insert(Files).values(host=data["host"],
                                               file=data["file"], hash=data["hash"],
                                               atime=data["atime"], ctime=data["ctime"],
                                               mtime=data["mtime"], size=data["size"]
                                               )
            self.session.execute(delete_stmt)
            self.session.execute(insert_stmt)
            self.session.commit()
        except (OperationalError, ResourceClosedError):
            logging.error("Failed to upsert on %s ((OperationalError,ResourceClosedError))" % self.driver_name)
            logging.error("Rolling back.")
            self.session.rollback()
            raise DrobloConnectivityLostException()
        except DrobloBroadException:
            logging.error("Failed to upsert on %s (DrobloBroadException)" % self.driver_name)
            logging.error("Rolling back.")
            self.session.rollback()
            raise DrobloConnectivityLostException()

    def delete_from_list(self, list_files_to_delete):
        if self.session:
            for data in list_files_to_delete:
                self._delete(data=data)

    def _delete(self, data):
        """
        _delete entries from files
        :param data:
        """
        if self.driver_name != __PSYCOPG2__:
            try:
                insert_stmt = insert(DeletedFiles). \
                    values(host=data["host"], file=data["deleted"])
                self.session.execute(insert_stmt)
                self.session.commit()
            except (OperationalError, ResourceClosedError):
                logging.error("Failed to delete on %s ((OperationalError,ResourceClosedError))" % self.driver_name)
                logging.error("Rolling back.")
                self.session.rollback()
                raise DrobloConnectivityLostException()
            except DrobloBroadException:
                logging.error("Failed to delete on %s (DrobloBroadException)" % self.driver_name)
                logging.error("Rolling back.")
                self.session.rollback()
                raise DrobloConnectivityLostException()

        try:
            delete_stmt = delete(Files). \
                where(Files.host == data["host"]). \
                where(Files.file.like(data["deleted"] + "%"))
            self.session.execute(delete_stmt)
            self.session.commit()
        except (OperationalError, ResourceClosedError):
            logging.error("Failed to delete on %s ((OperationalError,ResourceClosedError))" % self.driver_name)
            logging.error("Rolling back.")
            self.session.rollback()
            raise DrobloConnectivityLostException()
        except DrobloBroadException:
            logging.error("Failed to delete on %s (DrobloBroadException)" % self.driver_name)
            logging.error("Rolling back.")
            self.session.rollback()
            raise DrobloConnectivityLostException()

    def _regex_update(self, data):

        """
        insert or update into files
        :param data:
        """
        try:
            regex_stmt = update(Files). \
                where(Files.host == data["host"]). \
                values(file=func.regexp_replace(Files.file, data["src"], data["dst"]))
            self.session.execute(regex_stmt)
            self.session.commit()
        except (OperationalError, ResourceClosedError):
            logging.error("Failed to update on %s ((OperationalError,ResourceClosedError))" % self.driver_name)
            logging.error("Rolling back.")
            self.session.rollback()
            raise DrobloConnectivityLostException()
        except DrobloBroadException:
            logging.error("Failed to update on %s (DrobloBroadException)" % self.driver_name)
            logging.error("Rolling back.")
            self.session.rollback()
            raise DrobloConnectivityLostException()

    def _sqlite_update(self, data):
        """
        insert or update into files
        :param data:
        """
        try:
            original = self.session.query(Files). \
                filter(Files.host == data["host"]). \
                filter(Files.file == data["src"]).all()[0]

            delete_stmt = delete(Files).where(Files.host == data["host"]).where(Files.file == data["src"])

            insert_stmt = insert(Files).values(host=data["host"],
                                               file=data["dst"], hash=original.hash,
                                               atime=original.atime, ctime=original.ctime,
                                               mtime=original.mtime, size=original.size
                                               )
            self.session.execute(delete_stmt)
            self.session.execute(insert_stmt)
            self.session.commit()
        except (OperationalError, ResourceClosedError):
            logging.error("Failed to update on %s ((OperationalError,ResourceClosedError))" % self.driver_name)
            logging.error("Rolling back.")
            self.session.rollback()
            raise DrobloConnectivityLostException()
        except DrobloBroadException:
            logging.error("Failed to update on %s (DrobloBroadException)" % self.driver_name)
            logging.error("Rolling back.")
            self.session.rollback()
            raise DrobloConnectivityLostException()



# coding=utf-8
import setuptools

with open("README.md") as fh:
    long_description = fh.read()
setuptools.setup(
    name="droblo",
    version="0.0.8",
    author="KurisuD",
    author_email="KurisuD@pypi.darnand.net",
    # description="A network oriented, PostgreSQL backed, filesystem duplication detection and data mirroring system.",
    description="A multi-hosts, PostgreSQL backed, filesystem duplication detection system.",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/KurisuD/droblo",
    packages=setuptools.find_packages(),
    install_requires=['pap_logger', 'watchdog', 'pathlib', 'psycopg2-binary', 'SQLAlchemy'],
    python_requires='>=3.5',
    include_package_data=True,
    classifiers=[
        "Development Status :: 3 - Alpha",
        "Programming Language :: Python :: 3 :: Only",
        "Programming Language :: Python :: 3.5",
        "Intended Audience :: End Users/Desktop",
        "License :: OSI Approved :: GNU Affero General Public License v3 or later (AGPLv3+)",
        "Operating System :: OS Independent",
        "Topic :: Utilities",
        "Topic :: System :: Filesystems",
        # "Topic :: System :: Archiving :: Mirroring"
    ],
)
